/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is KaiRo.at Mandelbrot, XULRunner version.
 *
 * The Initial Developer of the Original Code is
 * Robert Kaiser <kairo@kairo.at>.
 * Portions created by the Initial Developer are Copyright (C) 2008-2011
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Robert Kaiser <kairo@kairo.at>
 *   prefiks (patch for some speedups)
 *   Boris Zbarsky <bzbarsky@mit.edu> (use imageData for canvas interaction)
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("resource://gre/modules/Services.jsm");

var gColorPalette = [];
var gStartTime = 0;
var gMbrotBundle;
var gCurrentImageData;

function Startup() {
  gMbrotBundle = document.getElementById("mbrotBundle");
  document.getElementById("statusLabel").value = gMbrotBundle.getString("statusEmpty");

  let img = document.getElementById("mbrotImage");
  img.addEventListener("mouseup", imgEvHandler, false);
  img.addEventListener("mousedown", imgEvHandler, false);
  img.addEventListener("mousemove", imgEvHandler, false);
  img.addEventListener("touchstart", imgEvHandler, false);
  img.addEventListener("touchend", imgEvHandler, false);
  img.addEventListener("touchcancel", imgEvHandler, false);
  img.addEventListener("touchleave", imgEvHandler, false);
  img.addEventListener("touchmove", imgEvHandler, false);

  Services.obs.notifyObservers(window, "mandelbrot-loaded", null);
}

function getAdjustPref(prefname) {
  let value;
  switch (prefname) {
    case "image.width":
    case "image.height":
      value = 0;
      try {
        value = Services.prefs.getIntPref("mandelbrot." + prefname);
      }
      catch (e) { }
      if ((value < 10) || (value > 5000)) {
        value = 300;
        Services.prefs.setIntPref("mandelbrot." + prefname, value);
      }
      return value;
    case "last_image.Cr_*":
      let Cr_min = -2.0;
      let Cr_max = 1.0;
      try {
        Cr_min = parseFloat(Services.prefs.getCharPref("mandelbrot.last_image.Cr_min"));
        Cr_max = parseFloat(Services.prefs.getCharPref("mandelbrot.last_image.Cr_max"));
      }
      catch (e) { }
      if ((Cr_min < -3) || (Cr_min > 2) ||
          (Cr_max < -3) || (Cr_max > 2) || (Cr_min >= Cr_max)) {
        Cr_min = -2.0; Cr_max = 1.0;
      }
      Services.prefs.setCharPref("mandelbrot.last_image.Cr_min", Cr_min);
      Services.prefs.setCharPref("mandelbrot.last_image.Cr_max", Cr_max);
      return {Cr_min: Cr_min, Cr_max: Cr_max};
    case "last_image.Ci_*":
      let Ci_min = -1.5;
      let Ci_max = 1.5;
      try {
        Ci_min = parseFloat(Services.prefs.getCharPref("mandelbrot.last_image.Ci_min"));
        Ci_max = parseFloat(Services.prefs.getCharPref("mandelbrot.last_image.Ci_max"));
      }
      catch (e) { }
      if ((Ci_min < -2.5) || (Ci_min > 2.5) ||
          (Ci_max < -2.5) || (Ci_max > 2.5) || (Ci_min >= Ci_max)) {
        Ci_min = -1.5; Ci_max = 1.5;
      }
      Services.prefs.setCharPref("mandelbrot.last_image.Ci_min", Ci_min);
      Services.prefs.setCharPref("mandelbrot.last_image.Ci_max", Ci_max);
      return {Ci_min: Ci_min, Ci_max: Ci_max};
    case "iteration_max":
      value = 500;
      try {
        value = Services.prefs.getIntPref("mandelbrot." + prefname);
      }
      catch (e) {
        setIter(value);
      }
      if (value < 10 || value > 10000) {
        value = 500;
        setIter(value);
      }
      return value;
    case "use_algorithm":
      value = "numeric";
      try {
        value = Services.prefs.getCharPref("mandelbrot." + prefname);
      }
      catch (e) {
        setAlgorithm(value);
      }
      return value;
   case "color_palette":
      value = "kairo";
      try {
        value = Services.prefs.getCharPref("mandelbrot." + prefname);
      }
      catch(e) {
        setPalette(value);
      }
      return value;
   case "syncProportions":
      value = true;
      try {
        value = Services.prefs.getBoolPref("mandelbrot." + prefname);
      }
      catch(e) {
        Services.prefs.setBoolPref("mandelbrot." + prefname, value);
      }
      return value;
    default:
      return false;
  }
}

function adjustCoordsAndDraw(aC_min, aC_max) {
  let iWidth = getAdjustPref("image.width");
  let iHeight = getAdjustPref("image.height");

  // correct coordinates
  if (aC_min.r < -2)
    aC_min.r = -2;
  if (aC_max.r > 2)
    aC_max.r = 2;
  if ((aC_min.r > 2) || (aC_max.r < -2) || (aC_min.r >= aC_max.r)) {
    aC_min.r = -2.0; aC_max.r = 1.0;
  }
  if (aC_min.i < -2)
    aC_min.i = -2;
  if (aC_max.i > 2)
    aC_max.i = 2;
  if ((aC_min.i > 2) || (aC_max.i < -2) || (aC_min.i >= aC_max.i)) {
    aC_min.i = -1.3; aC_max.i = 1.3;
  }

  let CWidth = aC_max.r - aC_min.r;
  let CHeight = aC_max.i - aC_min.i;
  let C_mid = new complex(aC_min.r + CWidth / 2, aC_min.i + CHeight / 2);

  let CRatio = Math.max(CWidth / iWidth, CHeight / iHeight);

  Services.prefs.setCharPref("mandelbrot.last_image.Cr_min", C_mid.r - iWidth * CRatio / 2);
  Services.prefs.setCharPref("mandelbrot.last_image.Cr_max", C_mid.r + iWidth * CRatio / 2);
  Services.prefs.setCharPref("mandelbrot.last_image.Ci_min", C_mid.i - iHeight * CRatio / 2);
  Services.prefs.setCharPref("mandelbrot.last_image.Ci_max", C_mid.i + iHeight * CRatio / 2);

  drawImage();
}

function drawImage() {
  let canvas = document.getElementById("mbrotImage");
  let context = canvas.getContext("2d");

  document.getElementById("drawButton").hidden = true;

  document.getElementById("statusLabel").value = gMbrotBundle.getString("statusDrawing");

  let Cr_vals = getAdjustPref("last_image.Cr_*");
  let Cr_min = Cr_vals.Cr_min;
  let Cr_max = Cr_vals.Cr_max;

  let Ci_vals = getAdjustPref("last_image.Ci_*");
  let Ci_min = Ci_vals.Ci_min;
  let Ci_max = Ci_vals.Ci_max;

  let iterMax = getAdjustPref("iteration_max");
  let algorithm = getAdjustPref("use_algorithm");

  let iWidth = getAdjustPref("image.width");
  let iHeight = getAdjustPref("image.height");

  gCurrentImageData = {
    C_min: new complex(Cr_min, Ci_min),
    C_max: new complex(Cr_max, Ci_max),
    iWidth: iWidth,
    iHeight: iHeight,
    iterMax: iterMax
  };

  canvas.width = iWidth;
  canvas.height = iHeight;

  context.fillStyle = "rgba(255, 255, 255, 127)";
  context.fillRect(0, 0, canvas.width, canvas.height);

  gStartTime = new Date();

  drawLine(0, [Cr_min, Cr_max, Ci_min, Ci_max],
              canvas, context, iterMax, algorithm);
}

function drawLine(line, dimensions, canvas, context, iterMax, algorithm) {
  let Cr_min = dimensions[0];
  let Cr_max = dimensions[1];
  let Cr_scale = Cr_max - Cr_min;

  let Ci_min = dimensions[2];
  let Ci_max = dimensions[3];
  let Ci_scale = Ci_max - Ci_min;

  let lines = Math.min(canvas.height - line, 8);
  let imageData = context.createImageData(canvas.width, lines);
  let pixels = imageData.data;
  let idx = 0;
  for (var img_y = line; img_y < canvas.height && img_y < line+8; img_y++)
    for (let img_x = 0; img_x < canvas.width; img_x++) {
      let C = new complex(Cr_min + (img_x / canvas.width) * Cr_scale,
                          Ci_min + (img_y / canvas.height) * Ci_scale);
      let colors = drawPoint(context, img_x, img_y, C, iterMax, algorithm);
      pixels[idx++] = colors[0];
      pixels[idx++] = colors[1];
      pixels[idx++] = colors[2];
      pixels[idx++] = colors[3];
    }
  context.putImageData(imageData, 0, line);

  if (img_y < canvas.height)
    setTimeout(drawLine, 0, img_y, dimensions, canvas, context, iterMax, algorithm);
  else if (gStartTime)
    EndCalc();
}

function EndCalc() {
  let endTime = new Date();
  let timeUsed = (endTime.getTime() - gStartTime.getTime()) / 1000;
  document.getElementById("statusLabel").value =
      gMbrotBundle.getFormattedString("statusTime", [timeUsed.toFixed(3)]);
  gStartTime = 0;
}

function complex(aReal, aImag) {
  this.r = aReal;
  this.i = aImag;
}
complex.prototype = {
  square: function() {
    return new complex(this.r * this.r - this.i * this.i,
                       2 * this.r * this.i);
  },
  dist: function() {
    return Math.sqrt(this.r * this.r + this.i * this.i);
  },
  add: function(aComplex) {
    return new complex(this.r + aComplex.r, this.i + aComplex.i);
  }
}

function mandelbrotValueOO (aC, aIterMax) {
  // this would be nice code in general but it looks like JS objects are too heavy for normal use.
  let Z = new complex(0.0, 0.0);
  for (var iter = 0; iter < aIterMax; iter++) {
    Z = Z.square().add(aC);
    if (Z.r * Z.r + Z.i * Z.i > 256) { break; }
  }
  return iter;
}

function mandelbrotValueNumeric (aC, aIterMax) {
  // optimized numeric code for fast calculation
  let Cr = aC.r, Ci = aC.i;
  let Zr = 0.0, Zi = 0.0;
  let Zr2 = Zr * Zr, Zi2 = Zi * Zi;
  for (var iter = 0; iter < aIterMax; iter++) {
    Zi = 2 * Zr * Zi + Ci;
    Zr = Zr2 - Zi2 + Cr;

    Zr2 = Zr * Zr; Zi2 = Zi * Zi;
    if (Zr2 + Zi2 > 256) { break; }
  }
  return iter;
}

function getColor(aIterValue, aIterMax) {
  let standardizedValue = Math.round(aIterValue * 1024 / aIterMax);
  if (gColorPalette && gColorPalette.length)
    return gColorPalette[standardizedValue];

  // fallback to simple b/w if for some reason we don't have a palette
  if (aIterValue == aIterMax)
    return [0, 0, 0, 255];
  else
    return [255, 255, 255, 255];
}

function getColorPalette(palName) {
  var palette = [];
  switch (palName) {
    case 'bw':
      for (let i = 0; i < 1024; i++) {
        palette[i] = [255, 255, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
    case 'kairo':
      // outer areas
      for (let i = 0; i < 32; i++) {
        let cc1 = Math.floor(i * 127 / 31);
        let cc2 = 170 - Math.floor(i * 43 / 31);
        palette[i] = [cc1, cc2, cc1, 255];
      }
      // inner areas
      for (let i = 0; i < 51; i++) {
        let cc = Math.floor(i * 170 / 50);
        palette[32 + i] = [cc, 0, (170-cc), 255];
      }
      // corona
      for (let i = 0; i < 101; i++) {
        let cc = Math.floor(i * 200 / 100);
        palette[83 + i] = [255, cc, 0, 255];
      }
      // inner corona
      for (let i = 0; i < 201; i++) {
        let cc1 = 255 - Math.floor(i * 85 / 200);
        let cc2 = 200 - Math.floor(i * 30 / 200);
        let cc3 = Math.floor(i * 170 / 200);
        palette[184 + i] = [cc1, cc2, cc3, 255];
      }
      for (let i = 0; i < 301; i++) {
        let cc1 = 170 - Math.floor(i * 43 / 300);
        let cc2 = 170 + Math.floor(i * 85 / 300);
        palette[385 + i] = [cc1, cc1, cc2, 255];
      }
      for (let i = 0; i < 338; i++) {
        let cc = 127 + Math.floor(i * 128 / 337);
        palette[686 + i] = [cc, cc, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
    case 'rainbow-linear1':
      for (let i = 0; i < 256; i++) {
        palette[i] = [i, 0, 0, 255];
        palette[256 + i] = [255, i, 0, 255];
        palette[512 + i] = [255 - i, 255, i, 255];
        palette[768 + i] = [i, 255-i, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
    case 'rainbow-squared1':
      for (let i = 0; i < 34; i++) {
        let cc = Math.floor(i * 255 / 33);
        palette[i] = [cc, 0, 0, 255];
      }
      for (let i = 0; i < 137; i++) {
        let cc = Math.floor(i * 255 / 136);
        palette[34 + i] = [255, cc, 0, 255];
      }
      for (let i = 0; i < 307; i++) {
        let cc = Math.floor(i * 255 / 306);
        palette[171 + i] = [255 - cc, 255, cc, 255];
      }
      for (let i = 0; i < 546; i++) {
        let cc = Math.floor(i * 255 / 545);
        palette[478 + i] = [cc, 255 - cc, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
    case 'rainbow-linear2':
      for (let i = 0; i < 205; i++) {
        let cc = Math.floor(i * 255 / 204);
        palette[i] = [255, cc, 0, 255];
        palette[204 + i] = [255 - cc, 255, 0, 255];
        palette[409 + i] = [0, 255, cc, 255];
        palette[614 + i] = [0, 255 - cc, 255, 255];
        palette[819 + i] = [cc, 0, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
    case 'rainbow-squared2':
      for (let i = 0; i < 19; i++) {
        let cc = Math.floor(i * 255 / 18);
        palette[i] = [255, cc, 0, 255];
      }
      for (let i = 0; i < 74; i++) {
        let cc = Math.floor(i * 255 / 73);
        palette[19 + i] = [255 - cc, 255, 0, 255];
      }
      for (let i = 0; i < 168; i++) {
        let cc = Math.floor(i * 255 / 167);
        palette[93 + i] = [0, 255, cc, 255];
      }
      for (let i = 0; i < 298; i++) {
        let cc = Math.floor(i * 255 / 297);
        palette[261 + i] = [0, 255 - cc, 255, 255];
      }
      for (let i = 0; i < 465; i++) {
        let cc = Math.floor(i * 255 / 464);
        palette[559 + i] = [cc, 0, 255, 255];
      }
      palette[1024] = [0, 0, 0, 255];
      break;
  }
  /*
     'Standard-Palette (QB-Colors)
     For i = 0 To 1024
         xx = CInt(i * 500 / 1024 + 2)
         If xx <= 15 Then clr = xx
         If xx > 15 Then clr = CInt(Sqr((xx - 15 + 1) * 15 ^ 2 / 485))
         If xx >= 500 Then clr = 0
         palette(i) = QBColor(clr)
     Next
  */
  return palette;
}

function drawPoint(context, img_x, img_y, C, iterMax, algorithm) {
  var itVal;
  switch (algorithm) {
    case 'oo':
      itVal = mandelbrotValueOO(C, iterMax);
      break;
    case 'numeric':
    default:
      itVal = mandelbrotValueNumeric(C, iterMax);
      break;
  }
  return getColor(itVal, iterMax);
}

/***** pure UI functions *****/

var zoomstart;
var imgBackup;

let imgEvHandler = {
  handleEvent: function(aEvent) {
    let canvas = document.getElementById("mbrotImage");
    let context = canvas.getContext("2d");
    switch (aEvent.type) {
      case 'mousedown':
      case 'touchstart':
        if (aEvent.button == 0) {
          // left button - start dragzoom
          zoomstart = {x: aEvent.clientX - canvas.offsetLeft,
                       y: aEvent.clientY - canvas.offsetTop};
          imgBackup = context.getImageData(0, 0, canvas.width, canvas.height);
        }
        break;
      case 'mouseup':
      case 'touchend':
        if (aEvent.button == 0 && zoomstart) {
          context.putImageData(imgBackup, 0, 0);
          let zoomend = {x: aEvent.clientX - canvas.offsetLeft,
                        y: aEvent.clientY - canvas.offsetTop};

          // make sure zoomend is bigger than zoomstart
          if ((zoomend.x == zoomstart.x) || (zoomend.y == zoomstart.y)) {
            // cannot zoom what has no area, discard it
            zoomstart = undefined;
            return;
          }
          if (zoomend.x < zoomstart.x)
            [zoomend.x, zoomstart.x] = [zoomstart.x, zoomend.x];
          if (zoomend.y < zoomstart.y)
            [zoomend.y, zoomstart.y] = [zoomstart.y, zoomend.y];

          // determine new "coordinates"
          let CWidth = gCurrentImageData.C_max.r - gCurrentImageData.C_min.r;
          let CHeight = gCurrentImageData.C_max.i - gCurrentImageData.C_min.i;
          let newC_min = new complex(
              gCurrentImageData.C_min.r + zoomstart.x / gCurrentImageData.iWidth * CWidth,
              gCurrentImageData.C_min.i + zoomstart.y / gCurrentImageData.iHeight * CHeight);
          let newC_max = new complex(
              gCurrentImageData.C_min.r + zoomend.x / gCurrentImageData.iWidth * CWidth,
              gCurrentImageData.C_min.i + zoomend.y / gCurrentImageData.iHeight * CHeight);

          adjustCoordsAndDraw(newC_min, newC_max);
        }
        zoomstart = undefined;
        break;
      case 'mousemove':
      case 'touchmove':
        if (aEvent.button == 0 && zoomstart) {
          context.putImageData(imgBackup, 0, 0);
          context.strokeStyle = "rgb(255,255,31)";
          context.strokeRect(zoomstart.x, zoomstart.y,
                             aEvent.clientX - canvas.offsetLeft - zoomstart.x,
                             aEvent.clientY - canvas.offsetTop - zoomstart.y);
        }
      break;
    }
  }
};

function saveImage() {
  const nsIFilePicker = Components.interfaces.nsIFilePicker;
  let fp = null;
  try {
    fp = Components.classes["@mozilla.org/filepicker;1"]
                   .createInstance(nsIFilePicker);
  } catch (e) {}
  if (!fp) return;
  let promptString = gMbrotBundle.getString("savePrompt");
  fp.init(window, promptString, nsIFilePicker.modeSave);
  fp.appendFilter(gMbrotBundle.getString("pngFilterName"), "*.png");
  fp.defaultString = "mandelbrot.png";

  let fpResult = fp.show();
  if (fpResult != nsIFilePicker.returnCancel) {
    saveCanvas(document.getElementById("mbrotImage"), fp.file);
  }
}

function exitMandelbrot() {
  var appInfo = Components.classes["@mozilla.org/xre/app-info;1"]
                          .getService(Components.interfaces.nsIXULAppInfo);
  if (appInfo.ID == "mandelbrot@kairo.at")
    quitApp(false);
  else
    window.close();
}

function updateBookmarkMenu(aParent) {
  document.getElementById("bookmarkSave").disabled =
    (!document.getElementById("drawButton").hidden || (gStartTime > 0));

  while (aParent.hasChildNodes() &&
         aParent.lastChild.id != "bookmarkSeparator")
    aParent.removeChild(aParent.lastChild);

  let file = Components.classes["@mozilla.org/file/directory_service;1"]
                       .getService(Components.interfaces.nsIProperties)
                       .get("ProfD", Components.interfaces.nsIFile);
  file.append("mandelbookmarks.sqlite");
  if (file.exists()) {
    let connection = Components.classes["@mozilla.org/storage/service;1"]
                               .getService(Components.interfaces.mozIStorageService)
                               .openDatabase(file);
    try {
      if (connection.tableExists("bookmarks")) {
        let statement = connection.createStatement(
            "SELECT name,ROWID FROM bookmarks ORDER BY ROWID ASC");
        while (statement.executeStep()) {
          let newItem = aParent.appendChild(document.createElement("menuitem"));
          newItem.setAttribute("label", statement.getString(0));
          newItem.setAttribute("bmRowID", statement.getString(1));
        }
        statement.reset();
        statement.finalize();
        return;
      }
    } finally {
      connection.close();
    }
  }
  // Create the "Nothing Available" Menu item and disable it.
  let na = aParent.appendChild(document.createElement("menuitem"));
  na.setAttribute("label", gMbrotBundle.getString("noBookmarks"));
  na.setAttribute("disabled", "true");
}

function callBookmark(evtarget) {
  if (evtarget.id == "bookmarkSave" || evtarget.id == "bookmarkSeparator")
    return;
  if (evtarget.id == "bookmarkOverview") {
    adjustCoordsAndDraw(new complex(0,0), new complex(0,0));
    return;
  }

  if (evtarget.getAttribute('bmRowID')) {
    let iterMax = 0;
    let C_min = null;
    let C_max = null;

    let file = Components.classes["@mozilla.org/file/directory_service;1"]
                         .getService(Components.interfaces.nsIProperties)
                         .get("ProfD", Components.interfaces.nsIFile);
    file.append("mandelbookmarks.sqlite");
    let connection = Components.classes["@mozilla.org/storage/service;1"]
                               .getService(Components.interfaces.mozIStorageService)
                               .openDatabase(file);
    let statement = connection.createStatement(
        "SELECT iteration_max,Cr_min,Cr_max,Ci_min,Ci_max FROM bookmarks WHERE ROWID=?1");
    statement.bindStringParameter(0, evtarget.getAttribute('bmRowID'));
    while (statement.executeStep()) {
      iterMax = statement.getInt32(0);
      C_min = new complex(statement.getDouble(1), statement.getDouble(3));
      C_max = new complex(statement.getDouble(2), statement.getDouble(4));
    }
    statement.finalize();
    connection.close();

    if (iterMax && C_min && C_max) {
      Services.prefs.setIntPref("mandelbrot.iteration_max", iterMax);
      adjustCoordsAndDraw(C_min, C_max);
    }
  }
}

function saveBookmark() {
  // retrieve wanted bookmark name with a prompt
  let prompts = Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                          .getService(Components.interfaces.nsIPromptService);
  let input = {value: ""}; // empty default value
  let ok = prompts.prompt(null, gMbrotBundle.getString("saveBookmarkTitle"), gMbrotBundle.getString("saveBookmarkLabel"), input, null, {});
  // ok is true if OK is pressed, false if Cancel. input.value holds the value of the edit field if "OK" was pressed.
  if (!ok || !input.value)
    return

  let bmName = input.value;

  // Open or create the bookmarks database.
  let file = Components.classes["@mozilla.org/file/directory_service;1"]
                       .getService(Components.interfaces.nsIProperties)
                       .get("ProfD", Components.interfaces.nsIFile);
  file.append("mandelbookmarks.sqlite");
  let connection = Components.classes["@mozilla.org/storage/service;1"]
                             .getService(Components.interfaces.mozIStorageService)
                             .openDatabase(file);
  connection.beginTransaction();
  if (!connection.tableExists("bookmarks"))
    connection.createTable("bookmarks", "name TEXT, iteration_max INTEGER, Cr_min REAL, Cr_max REAL, Ci_min REAL, Ci_max REAL");
  // NULL. The value is a NULL value.
  // INTEGER. The value is a signed integer, stored in 1, 2, 3, 4, 6, or 8 bytes depending on the magnitude of the value.
  // REAL. The value is a floating point value, stored as an 8-byte IEEE floating point number.
  // TEXT. The value is a text string, stored using the database encoding (UTF-8, UTF-16BE or UTF-16-LE).

  // Put value of the current image into the bookmarks table
  let statement = connection.createStatement(
      "INSERT INTO bookmarks (name,iteration_max,Cr_min,Cr_max,Ci_min,Ci_max) VALUES (?1,?2,?3,?4,?5,?6)");
  statement.bindStringParameter(0, bmName);
  statement.bindStringParameter(1, gCurrentImageData.iterMax);
  statement.bindStringParameter(2, gCurrentImageData.C_min.r);
  statement.bindStringParameter(3, gCurrentImageData.C_max.r);
  statement.bindStringParameter(4, gCurrentImageData.C_min.i);
  statement.bindStringParameter(5, gCurrentImageData.C_max.i);
  statement.execute();
  statement.finalize();
  connection.commitTransaction();
  connection.close();
}

function imgSettings() {
  let anchor = null;
  let position = "before_start";
  if (document.getElementById("mandelbrot-page").nodeName == "page") {
    anchor = document.getElementById("mandelbrotToolbar");
  }
  else {
    anchor = document.getElementById("mandelbrotMenubar");
    position = "after_start";
  }
  document.getElementById("imgSettingsPanel").showPopup(anchor, position);
}

function initImgSettings() {
  // Get values from prefs.
  for each (let coord in ["Cr", "Ci"]) {
    let coord_vals = getAdjustPref("last_image." + coord + "_*");
    document.getElementById("is_" + coord + "_min").value = coord_vals[coord + "_min"];
    document.getElementById("is_" + coord + "_max").value = coord_vals[coord + "_max"];
  }
  for each (let dim in ["width", "height"]) {
    document.getElementById("is_img_" + dim).value = getAdjustPref("image." + dim);
  }
  document.getElementById("is_syncProp").checked = getAdjustPref("syncProportions");

  // Calculate scales.
  recalcCoord("Cr", "scale");
  recalcCoord("Ci", "scale");

  // Clear the preview.
  let canvas = document.getElementById("is_mbrotPreview");
  let context = canvas.getContext("2d");
  context.fillStyle = "rgba(255, 255, 255, 127)";
  context.fillRect(0, 0, canvas.width, canvas.height);

  // Set lists to correct values.
  updateIterList();
  updatePaletteList();
  updateAlgoList();
}

function closeImgSettings() {
  // Hide popup, which will automatically make a call to save values.
  document.getElementById("imgSettingsPanel").hidePopup();
}

function saveImgSettings() {
  // Get values to prefs.
  for each (let coord in ["Cr_min", "Cr_max", "Ci_min", "Ci_max"]) {
    Services.prefs.setCharPref("mandelbrot.last_image." + coord,
                               document.getElementById("is_" + coord).value);
  }
  for each (let dim in ["width", "height"]) {
    Services.prefs.setIntPref("mandelbrot.image." + dim,
                              document.getElementById("is_img_" + dim).value);
  }
  Services.prefs.setBoolPref("mandelbrot.syncProportions",
                             document.getElementById("is_syncProp").checked);
}

function checkISValue(textbox, type) {
  if (type == "coord") {
    textbox.value = roundCoord(parseFloat(textbox.value));
  }
  else if (type == "dim") {
    textbox.value = parseInt(textbox.value);
  }
}

function drawPreview() {
  let canvas = document.getElementById("is_mbrotPreview");
  let context = canvas.getContext("2d");

  if (document.getElementById("is_img_width").value /
      document.getElementById("is_img_height").value
        < 80 / 50) {
    canvas.height = 50;
    canvas.width = canvas.height *
      document.getElementById("is_img_width").value /
      document.getElementById("is_img_height").value;
  }
  else {
    canvas.width = 80;
    canvas.height = canvas.width *
      document.getElementById("is_imgHeight").value /
      document.getElementById("is_imgWidth").value;
  }

  let Cr_min = parseFloat(document.getElementById("is_Cr_min").value);
  let Cr_max = parseFloat(document.getElementById("is_Cr_max").value);
  if ((Cr_min < -2) || (Cr_min > 2) ||
      (Cr_max < -2) || (Cr_max > 2) || (Cr_min >= Cr_max)) {
    Cr_min = -2.0; Cr_max = 1.0;
  }

  let Ci_min = parseFloat(document.getElementById("is_Ci_min").value);
  let Ci_max = parseFloat(document.getElementById("is_Ci_max").value);
  if ((Ci_min < -2) || (Ci_min > 2) ||
      (Ci_max < -2) || (Ci_max > 2) || (Ci_min >= Ci_max)) {
    Ci_min = -2.0; Ci_max = 1.0;
  }

  let iterMax = getAdjustPref("iteration_max");
  let algorithm = getAdjustPref("use_algorithm");

  context.fillStyle = "rgba(255, 255, 255, 127)";
  context.fillRect(0, 0, canvas.width, canvas.height);

  let currentPalette = getAdjustPref("color_palette");
  gColorPalette = getColorPalette(currentPalette);

  drawLine(0, [Cr_min, Cr_max, Ci_min, Ci_max],
              canvas, context, iterMax, algorithm);
}

function recalcCoord(coord, target) {
  let othercoord = (coord == "Ci") ? "Cr" : "Ci";
  let owndim = (coord == "Ci") ? "height" : "width";
  let otherdim = (coord == "Ci") ? "width" : "height";
  let myscale;
  if (target == "scale") {
    myscale =
      parseFloat(document.getElementById("is_" + coord + "_max").value) -
      parseFloat(document.getElementById("is_" + coord + "_min").value);
    document.getElementById("is_" + coord + "_scale").value = roundCoord(myscale);
  }
  else if (target == 'max') {
    let mymax =
      parseFloat(document.getElementById("is_" + coord + "_min").value) +
      parseFloat(document.getElementById("is_" + coord + "_scale").value);
    document.getElementById("is_" + coord + "_max").value = roundCoord(mymax);
    myscale = document.getElementById("is_" + coord + "_scale").value;
  }
  if (document.getElementById("is_syncProp").checked) {
    let otherscale = myscale *
      document.getElementById("is_img_" + otherdim).value /
      document.getElementById("is_img_" + owndim).value;
    document.getElementById("is_" + othercoord + "_scale").value = roundCoord(otherscale);
    let othermax =
      parseFloat(document.getElementById("is_" + othercoord + "_min").value) +
      parseFloat(document.getElementById("is_" + othercoord + "_scale").value);
    document.getElementById("is_" + othercoord + "_max").value = roundCoord(othermax);
  }
}

function checkProportions() {
  if (!document.getElementById("is_syncProp").checked) {
    recalcCoord("Cr", "scale");
  }
}

function roundCoord(floatval) {
  // We should round to 10 decimals here or so
  return parseFloat(floatval.toFixed(10));
}

function updateIterList() {
  let currentIter = getAdjustPref("iteration_max");
  document.getElementById("iterList").value = currentIter;
}

function updatePaletteList() {
  let currentPalette = getAdjustPref("color_palette");
  if (!gColorPalette || !gColorPalette.length)
    gColorPalette = getColorPalette(currentPalette);
  document.getElementById("colorList").value = currentPalette;
}

function updateAlgoList() {
  let currentAlgo = getAdjustPref("use_algorithm");
  document.getElementById("algoList").value = currentAlgo;
}

function setIter(aIter) {
  Services.prefs.setIntPref("mandelbrot.iteration_max", aIter);
}

function setPalette(aPaletteID) {
  Services.prefs.setCharPref("mandelbrot.color_palette", aPaletteID);
  gColorPalette = getColorPalette(aPaletteID);
}

function setAlgorithm(algoID) {
  Services.prefs.setCharPref("mandelbrot.use_algorithm", algoID);
}


/***** helper functions from external sources *****/

// function below is based on http://developer.mozilla.org/en/docs/Code_snippets:Canvas
// custom modifications:
//   - use "a"-prefix on function arguments
//   - take an nsILocalFile as aDestFile argument
//   - always do silent download
function saveCanvas(aCanvas, aDestFile) {
  // create a data url from the canvas and then create URIs of the source and targets  
  var io = Components.classes["@mozilla.org/network/io-service;1"]
                     .getService(Components.interfaces.nsIIOService);
  var source = io.newURI(aCanvas.toDataURL("image/png", ""), "UTF8", null);

  // prepare to save the canvas data
  var persist = Components.classes["@mozilla.org/embedding/browser/nsWebBrowserPersist;1"]
                          .createInstance(Components.interfaces.nsIWebBrowserPersist);

  persist.persistFlags = Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_REPLACE_EXISTING_FILES;
  persist.persistFlags |= Components.interfaces.nsIWebBrowserPersist.PERSIST_FLAGS_AUTODETECT_APPLY_CONVERSION;

  // save the canvas data to the file
  persist.saveURI(source, null, null, null, null, aDestFile);
}
